"""Module for calculating theoretical optimal defaults for SVRG and GLM configurations."""

import warnings
from functools import wraps
from typing import Any, Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from numpy.typing import NDArray

from .. import pytrees


def _convert_to_float(func):
    """
    Decorator to convert all inputs to float before passing them to the function.

    Ensures that calculations within the function are performed with floating-point precision.

    Parameters
    ----------
    func :
        The function to be wrapped by the decorator.

    Returns
    -------
    :
        Wrapped function with inputs converted to floats.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        args, kwargs = jax.tree_util.tree_map(float, (args, kwargs))
        return func(*args, **kwargs)

    return wrapper


def svrg_optimal_batch_and_stepsize(
    compute_smoothness_constants: Callable,
    *data: Any,
    batch_size: Optional[int] = None,
    stepsize: Optional[float] = None,
    strong_convexity: Optional[float] = None,
    n_power_iters: Optional[int] = 20,
    default_batch_size: int = 1,
    default_stepsize: float = 1e-3,
):
    """
    Calculate the optimal batch size and step size for SVRG optimization in GLMs.

    This function computes the optimal batch size and step size parameters for SVRG,
    based on the smoothness constants and strong convexity of the loss function.

    Parameters
    ----------
    compute_smoothness_constants :
        Function that computes the smoothness constants `l_smooth` and `l_smooth_max` for the problem.
        This is problem (loss function) specific.
    data :
        Input data, typically (X, y) for a GLM.
    batch_size :
        The batch size set by the user. If None, it will be calculated.
    stepsize :
        The step size set by the user. If None, it will be calculated.
    strong_convexity :
        The strong convexity constant. For L2-regularized losses, this should be the regularization strength.
    n_power_iters :
        Maximum number of iterations for the power method when finding the largest eigenvalue. If set to `None`,
        the lead eigenvalue is computed directly (no power method).
    default_batch_size :
        Default batch size to use if the optimal calculation fails.
    default_stepsize :
        Default step size to use if the optimal calculation fails.

    Returns
    -------
    dict
        Dictionary containing the optimal `batch_size` and `stepsize`.

    Raises
    ------
    ValueError
        If the data provided has inconsistent numbers of samples.

    Warnings
    --------
    UserWarning
        Warns the user if the calculation fails and defaults are used instead.

    Examples
    --------
    >>> import numpy as np
    >>> from nemos.solvers import svrg_optimal_batch_and_stepsize as compute_opt_params
    >>> from nemos.solvers import glm_softplus_poisson_l_max_and_l
    >>> np.random.seed(123)
    >>> X = np.random.normal(size=(500, 5))
    >>> y = np.random.poisson(np.exp(X.dot(np.ones(X.shape[1]))))
    >>> batch_size, stepsize = compute_opt_params(
    ...     glm_softplus_poisson_l_max_and_l, X, y, strong_convexity=0.08, batch_size=1
    ... )
    """
    # Get the number of samples, ensuring consistency across all inputs
    num_samples = {dd.shape[0] for dd in jax.tree_util.tree_leaves(data)}
    if len(num_samples) != 1:
        raise ValueError("Each array in data must have the same number of samples.")
    num_samples = num_samples.pop()

    # If both parameters are set by the user, return them directly
    if batch_size is not None and stepsize is not None:
        return {"batch_size": batch_size, "stepsize": stepsize}

    # Compute smoothness constants
    l_smooth_max, l_smooth = compute_smoothness_constants(
        *data, n_power_iters=n_power_iters, batch_size=batch_size
    )

    # Compute optimal batch size if not provided by the user
    if batch_size is None:
        batch_size = _calculate_optimal_batch_size_svrg(
            num_samples, l_smooth_max, l_smooth, strong_convexity=strong_convexity
        )

    # Fall back to defaults if batch size calculation fails
    if not jnp.isfinite(batch_size):
        batch_size = default_batch_size
        warnings.warn(
            "Could not determine batch size automatically. "
            f"Falling back on the default values of {default_batch_size}.",
            UserWarning,
        )

    # Compute optimal step size if not provided by the user
    if stepsize is None:
        stepsize = _calculate_stepsize_svrg(
            batch_size, num_samples, l_smooth_max, l_smooth
        )
        if stepsize < 0:
            stepsize = default_stepsize
            warnings.warn(
                "Could not determine step size automatically. "
                f"Falling back on the default value of {default_stepsize}.",
                UserWarning,
            )
    return {"batch_size": int(batch_size), "stepsize": stepsize}


def glm_softplus_poisson_l_max_and_l(
    *data: NDArray, n_power_iters: Optional[int] = 20, batch_size: Optional[int] = None
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    Calculate smoothness constants for a Poisson GLM with a softplus inverse link function.

    Computes the smoothness constant (`l_smooth`) and the maximum smoothness constant
    (`l_smooth_max`) for SVRG, given the data and the GLM structure.

    Parameters
    ----------
    data :
        Input data, typically (X, y).
    n_power_iters :
        Number of power iterations to use when finding the largest eigenvalue. If None,
        the eigenvalue is calculated directly.
    batch_size:
        The batch size set by the user. If None, it will be n_samples.

    Returns
    -------
    :
        Maximum smoothness constant (`l_smooth_max`) and smoothness constant (`l_smooth`).
    """
    X, y = data

    if batch_size is None:
        batch_size = y.shape[0]

    # takes care of population glm (see bound found on overleaf)
    y = jnp.max(y, axis=tuple(range(1, y.ndim)))

    # concatenate all data (if X is FeaturePytree)
    if isinstance(X, (pytrees.FeaturePytree, dict)):
        X = jnp.hstack(jax.tree_util.tree_leaves(X))

    l_smooth = _glm_softplus_poisson_l_smooth(
        X, y, n_power_iters=n_power_iters, batch_size=batch_size
    )
    l_smooth_max = _glm_softplus_poisson_l_smooth_max(X, y, batch_size=batch_size)
    return l_smooth_max, l_smooth


def _glm_softplus_poisson_l_smooth_multiply(
    X: NDArray, y: NDArray, v: NDArray, batch_size: int
):
    """
    Multiply vector `v` with the matrix X.T @ D @ X without forming it explicitly.

    This method estimates the multiplication by calculating the Hessian of the loss.
    It is efficient for situations where X can fit in memory.
    If batch_size is provided, the computation will be done by slicing the array.

    Parameters
    ----------
    X :
        Input data matrix (N x d).
    y :
        Output data vector (N,).
    v :
        Vector to be multiplied (d,).

    Returns
    -------
    :
        Result of the multiplication (X.T @ D @ X) @ v.
    """
    N, K = X.shape
    out = jnp.zeros((K,))
    for i in range(0, N, batch_size):
        xb, yb = X[i : i + batch_size], y[i : i + batch_size]
        out = out + xb.T.dot((0.17 * yb + 0.25) * xb.dot(v))
    out = out / N
    return out


def _glm_softplus_poisson_l_smooth_with_power_iteration(
    X: NDArray, y: NDArray, n_power_iters: int = 20, batch_size: Optional[int] = None
):
    """
    Compute the largest eigenvalue of X.T @ D @ X using the power method.

    Instead of calculating the full matrix and its eigenvalue directly, this function
    uses the power iteration method, which is more memory efficient and scales better
    with large datasets.

    Parameters
    ----------
    X :
        Input data matrix (N x d).
    y :
        Output data vector (N,).
    n_power_iters :
        Maximum number of power iterations to use.
    batch_size :
        The batch size, if user provides one.

    Returns
    -------
    :
        The largest eigenvalue of X.T @ D @ X.
    """

    if batch_size is None:
        batch_size = X.shape[0]

    _, d = X.shape

    # Initialize a random d-dimensional vector for power iteration
    v = jnp.ones((d,))

    # Run power iteration to approximate the largest eigenvalue
    for _ in range(n_power_iters):
        v_prev = v.copy()
        v = _glm_softplus_poisson_l_smooth_multiply(X, y, v, batch_size)
        v /= v.max()

        # Check for convergence
        if jnp.allclose(v_prev, v):
            break

    # Final eigenvalue calculation
    v /= jnp.linalg.norm(v)
    return _glm_softplus_poisson_l_smooth_multiply(X, y, v, batch_size).dot(v)


def _glm_softplus_poisson_l_smooth(
    X: NDArray, y: NDArray, batch_size: int, n_power_iters: Optional[int] = 20
) -> jnp.ndarray:
    """
    Calculate the smoothness constant `L` for a Poisson GLM with softplus inverse link.

    Depending on whether `n_power_iters` is provided, this function either computes
    the largest eigenvalue directly or uses the power method.

    Parameters
    ----------
    X :
        Input data matrix (N x d).
    y :
        Output data vector (N,).
    batch_size :
        The batch size, if user provides one.
    n_power_iters :
        Number of power iterations to use when finding the largest eigenvalue. If None,
        the eigenvalue is calculated directly.

    Returns
    -------
    :
        Smoothness constant `L`.

    Raises
    ------
    TypeError:
        If `n_power_iters` is not an integer or None.
    RuntimeError:
        If the eigienvalue computation does not converge.
    ValueError:
        If a negative integer is provided.
    """
    if n_power_iters and not isinstance(n_power_iters, int):
        raise TypeError("`n_power_iters` must be an integer or None.")

    if n_power_iters is None:
        try:
            warnings.warn(
                "Direct computation of the eigenvalues requires a substantial amount of resources. "
                "Please, consider using the power method by setting the `n_power_iters` parameter "
                "(default behavior).",
                UserWarning,
            )
            # Calculate the Hessian directly and find the largest eigenvalue
            XDX = X.T.dot((0.17 * y.reshape(y.shape[0], 1) + 0.25) * X) / y.shape[0]
            return jnp.sort(jnp.linalg.eigvalsh(XDX))[-1]
        except RuntimeError as e:
            raise RuntimeError(
                f"Failed to calculate the largest eigenvalue: {e}. "
                "Consider using the power method by setting the `n_power_iters` parameter."
            )
    else:
        if n_power_iters <= 0:
            raise ValueError("`n_power_iters` must be positive.")
        # Use power iteration to find the largest eigenvalue
        return _glm_softplus_poisson_l_smooth_with_power_iteration(
            X, y, n_power_iters, batch_size=batch_size
        )


def _glm_softplus_poisson_l_smooth_max(
    X: NDArray, y: NDArray, batch_size: int
) -> NDArray:
    """
    Calculate the maximum smoothness constant `L_max` for individual observations.

    This function estimates the maximum smoothness constant among the individual
    components of the loss function.

    Parameters
    ----------
    X :
        Input data matrix (N x d).
    y :
        Output data vector (N,).
    batch_size:
        The batch size.

    Returns
    -------
    l_max :
        Maximum smoothness constant `L_max`.
    """
    N, _ = X.shape

    l_max = jnp.array([0])
    for nb in range(0, N, batch_size):
        xb, yb = X[nb : nb + batch_size], y[nb : nb + batch_size]
        l_max = jnp.maximum(
            l_max, jnp.max(jnp.linalg.norm(xb, axis=1) ** 2 * (0.17 * yb + 0.25))
        )

    return l_max[0]


@_convert_to_float
def _calculate_stepsize_svrg(
    batch_size: int, num_samples: int, l_smooth_max: float, l_smooth: float
):
    """
    Calculate optimal step size for SVRG according to [1].

    Parameters
    ----------
    batch_size :
        Mini-batch size.
    num_samples :
        Overall number of data points.
    l_smooth_max :
        Maximum smoothness constant among f_{i}.
    l_smooth :
        Smoothness constant.

    Returns
    -------
    :
        Optimal step size for the optimization.

    References
    ----------
    [1] Sebbouh, Othmane, et al. "Towards closing the gap between the theory and practice of SVRG."
    Advances in neural information processing systems 32 (2019).
    """
    numerator = 0.5 * batch_size * (num_samples - 1)
    denominator = (
        3 * (num_samples - batch_size) * l_smooth_max
        + num_samples * (batch_size - 1) * l_smooth
    )
    return numerator / denominator


def _calculate_optimal_batch_size_svrg(
    num_samples: int,
    l_smooth_max: float,
    l_smooth: float,
    strong_convexity: Optional[float] = None,
) -> int:
    r"""
    Calculate the optimal batch size for SVRG based on theoretical guidelines.

    The batch size is computed according to the smoothness constants, strong convexity,
    and number of samples, following the recommendations in Table 1 of [1].

    Parameters
    ----------
    num_samples:
        The number of samples.
    l_smooth_max:
        The $L_{\text{max}}$ smoothness constant.
    l_smooth:
        The $L$ smoothness constant.
    strong_convexity:
        The strong convexity constant.

    Returns
    -------
    batch_size:
        The optimal mini-batch size for SVRG.

    References
    ----------
    [1] Sebbouh, Othmane, et al. "Towards closing the gap between the theory and practice of SVRG."
    Advances in neural information processing systems 32 (2019).
    """
    if strong_convexity is None:
        # Assume that num_sample is large enough for mini-batching.
        # This is usually the case for neuroscience where num_sample
        # is typically very large.
        # If this assumption is not matched, convergence may be slow.
        batch_size = 1
    else:
        # Compute optimal batch size according to Table 1.
        if num_samples >= 3 * l_smooth_max / strong_convexity:
            batch_size = 1
        elif num_samples > l_smooth / strong_convexity:
            b_tilde = _calculate_b_tilde(
                num_samples, l_smooth_max, l_smooth, strong_convexity
            )
            if l_smooth_max < num_samples * l_smooth / 3:
                b_hat = _calculate_b_hat(num_samples, l_smooth_max, l_smooth)
                batch_size = int(jnp.floor(jnp.minimum(b_hat, b_tilde)))
            else:
                batch_size = int(jnp.floor(b_tilde))
        else:
            if l_smooth_max < num_samples * l_smooth / 3:
                batch_size = int(
                    jnp.floor(_calculate_b_hat(num_samples, l_smooth_max, l_smooth))
                )
            else:
                batch_size = int(num_samples)  # reset this to int
    return batch_size


@_convert_to_float
def _calculate_b_hat(num_samples: int, l_smooth_max: float, l_smooth: float):
    r"""
    Calculate the optimal `b_hat` batch size parameter for SVRG.

    This is a helper function to compute the theoretical batch size $\hat{b}$, as detailed
    in "Table 1" of [1].

    Parameters
    ----------
    num_samples :
        Total number of data points.
    l_smooth_max :
        Maximum smoothness constant $L_{\text{max}}$.
    l_smooth :
        Smoothness constant $L$.

    Returns
    -------
    float
        Optimal batch size parameter `b_hat`.

    References
    ----------
    [1] Sebbouh, Othmane, et al. "Towards closing the gap between the theory and practice of SVRG."
    Advances in neural information processing systems 32 (2019).
    """
    numerator = num_samples / 2 * (3 * l_smooth_max - l_smooth)
    denominator = num_samples * l_smooth - 3 * l_smooth_max
    return jnp.sqrt(numerator / denominator)


@_convert_to_float
def _calculate_b_tilde(num_samples, l_smooth_max, l_smooth, strong_convexity):
    r"""
    Calculate the optimal $\tilde{b}$ batch size parameter for SVRG.

    This is a helper function to compute the theoretical batch size  $\tilde{b}$, as detailed
    in "Table 1" of [1].

    Parameters
    ----------
    num_samples :
        Total number of data points.
    l_smooth_max :
        Maximum smoothness constant `L_max`.
    l_smooth :
        Smoothness constant `L`.
    strong_convexity :
        Strong convexity constant.

    Returns
    -------
    :
        Optimal batch size parameter `b_tilde`.

    References
    ----------
    [1] Sebbouh, Othmane, et al. "Towards closing the gap between the theory and practice of SVRG."
    Advances in neural information processing systems 32 (2019).
    """
    numerator = (3 * l_smooth_max - l_smooth) * num_samples
    denominator = (
        num_samples * (num_samples - 1) * strong_convexity
        - num_samples * l_smooth
        + 3 * l_smooth_max
    )
    return numerator / denominator



from contextlib import nullcontext as does_not_raise

import jax
import numpy as np
import pynapple as nap
import pytest

from nemos import convolve, utils


class TestShiftTimeAxisAndConvolve:

    @pytest.mark.parametrize(
        "time_series, check_func, axis",
        [
            (np.zeros((1, 20)), lambda x: x.ndim == 3, 1),
            (np.zeros((20,)), lambda x: x.ndim == 2, 0),
            (np.zeros((20, 1)), lambda x: x.ndim == 3, 0),
            (np.zeros((1, 20, 1)), lambda x: x.ndim == 4, 0),
        ],
    )
    def test_output_ndim(self, time_series, check_func, axis):
        """Check that the output dimensionality matches expectation."""
        res = convolve._shift_time_axis_and_convolve(
            time_series, np.zeros((1, 1)), axis=axis
        )
        if not utils.pytree_map_and_reduce(check_func, all, res):
            raise ValueError("Output doesn't match expected structure")

    @pytest.mark.parametrize(
        "time_series, axis, output_shape",
        [
            (np.zeros((20, 1)), 0, (20, 1, 1)),
            (np.zeros((1, 20, 1)), 1, (1, 20, 1, 1)),
        ],
    )
    def test_output_shape(self, time_series, axis, output_shape):
        """Check that the output shape matches expectation."""

        def check_func(x):
            return x.shape == output_shape

        res = convolve._shift_time_axis_and_convolve(
            time_series, np.zeros((1, 1)), axis=axis
        )
        if not utils.pytree_map_and_reduce(check_func, all, res):
            raise ValueError("Output  number of neuron doesn't match input.")

    @pytest.mark.parametrize(
        "time_series, axis",
        [
            (np.zeros((20,)), 0),
            (np.zeros((20, 1)), 0),
            (np.zeros((1, 20, 1)), 1),
        ],
    )
    @pytest.mark.parametrize("basis_matrix", [np.zeros((1, 1)), np.zeros((1, 2))])
    def test_output_num_basis(self, time_series, basis_matrix, axis):
        """Check that the number of features in input and output matches."""

        def check_func(conv):
            return basis_matrix.shape[-1] == conv.shape[-1]

        res = convolve._shift_time_axis_and_convolve(
            time_series, basis_matrix, axis=axis
        )
        if not utils.pytree_map_and_reduce(check_func, all, res):
            raise ValueError("Output  number of neuron doesn't match input.")

    @pytest.mark.parametrize(
        "basis_matrix", [np.random.normal(size=(4, 3)) for _ in range(2)]
    )
    @pytest.mark.parametrize(
        "trial_counts", [np.random.normal(size=(2, 10, 3)) for _ in range(2)]
    )
    def test_valid_convolution_output(self, basis_matrix, trial_counts):
        """Check output matches numpy convolve."""
        numpy_out = np.zeros(
            (
                trial_counts.shape[0],
                trial_counts.shape[1] - basis_matrix.shape[0] + 1,
                trial_counts.shape[2],
                basis_matrix.shape[1],
            )
        )
        for tri_i, trial in enumerate(trial_counts):
            for neu_k, vec in enumerate(trial.T):
                for bas_j, basis in enumerate(basis_matrix.T):
                    numpy_out[tri_i, :, neu_k, bas_j] = np.convolve(
                        vec, basis, mode="valid"
                    )

        utils_out = np.asarray(
            convolve._shift_time_axis_and_convolve(trial_counts, basis_matrix, axis=1)
        )
        assert np.allclose(utils_out, numpy_out, rtol=10**-5, atol=10**-5), (
            "Output of utils.convolve_1d_trials "
            "does not match numpy.convolve in "
            '"valid" mode.'
        )


class TestCreateConvolutionalPredictor:

    @pytest.mark.parametrize("basis_matrix", [np.zeros((3,) * n) for n in [0, 1, 2, 3]])
    @pytest.mark.parametrize("trial_count_shape", [(1, 30, 2), (2, 10, 20)])
    def test_basis_number_of_dim(self, basis_matrix, trial_count_shape: tuple[int]):
        vec = np.ones(trial_count_shape)
        raise_exception = basis_matrix.ndim != 2
        if raise_exception:
            with pytest.raises(
                ValueError, match="basis_matrix must be a 2 dimensional"
            ):
                convolve.create_convolutional_predictor(basis_matrix, vec, axis=1)
        else:
            convolve.create_convolutional_predictor(basis_matrix, vec, axis=1)

    @pytest.mark.parametrize("basis_matrix", [np.zeros((3, 4))])
    @pytest.mark.parametrize(
        "trial_counts, expectation, axis",
        [
            (np.zeros((1, 30, 2)), does_not_raise(), 1),
            ([np.zeros((30, 2))], does_not_raise(), 0),
            ({"tr1": np.zeros((30, 2)), "tr2": np.zeros((30, 2))}, does_not_raise(), 0),
            (np.zeros((1, 30, 1, 2)), does_not_raise(), 1),
            (
                [np.array(10)],
                pytest.raises(
                    ValueError,
                    match="`time_series` should contain arrays of at least one",
                ),
                0,
            ),
            (np.zeros((30, 10)), does_not_raise(), 0),
            ([np.zeros((30, 10))], does_not_raise(), 1),
            (np.zeros(10), does_not_raise(), 0),
        ],
    )
    def test_spike_count_type(self, basis_matrix, expectation, trial_counts, axis):
        with expectation:
            convolve.create_convolutional_predictor(
                basis_matrix, trial_counts, axis=axis
            )

    @pytest.mark.parametrize("basis_matrix", [np.zeros((4, 3))])
    @pytest.mark.parametrize(
        "trial_counts",
        [
            np.zeros((1, 4, 2)),  # valid
            np.zeros((1, 40, 2)),  # valid
            np.zeros((1, 3, 2)),  # invalid
        ],
    )
    def test_sufficient_trial_duration(self, basis_matrix, trial_counts):
        raise_exception = trial_counts.shape[1] < basis_matrix.shape[0]
        if raise_exception:
            with pytest.raises(
                ValueError,
                match="Insufficient trial duration. The number of time points",
            ):
                convolve.create_convolutional_predictor(
                    basis_matrix, trial_counts, axis=1
                )
        else:
            convolve.create_convolutional_predictor(basis_matrix, trial_counts, axis=1)

    @pytest.mark.parametrize(
        "basis_matrix, expectation",
        [
            (np.zeros((2, 1)), does_not_raise()),
            (
                np.zeros((0, 1)),
                pytest.raises(
                    ValueError, match=r"Empty array provided. At least one of dimension"
                ),
            ),
        ],
    )
    def test_empty_basis(self, basis_matrix, expectation):
        vec = np.ones((1, 10))
        with expectation:
            convolve.create_convolutional_predictor(basis_matrix, vec, axis=1)

    @pytest.mark.parametrize("window_size", [2])
    @pytest.mark.parametrize("trial_len", [4, 5])
    @pytest.mark.parametrize("array_dim", [1, 2, 3])
    def test_output_trial_length(self, window_size, trial_len, array_dim):
        basis_matrix = np.zeros((window_size, 1))
        time_series = np.zeros((trial_len,))
        sample_axis = 0

        if array_dim == 2:
            time_series = np.expand_dims(time_series, axis=0)
            sample_axis = 1
        if array_dim == 3:
            time_series = np.expand_dims(time_series, axis=(0, 2))
            sample_axis = 1

        res = convolve.create_convolutional_predictor(
            basis_matrix, time_series, axis=sample_axis
        )
        if res.shape[sample_axis] != trial_len:
            raise ValueError(
                "The output of create_convolutional_predictor should have size num_samples!"
            )

    @pytest.mark.parametrize("basis_matrix", [np.zeros((4, 3))])
    @pytest.mark.parametrize(
        "trial_counts",
        [
            np.zeros((0, 4, 3)),  # invalid
            np.zeros((1, 40, 0)),  # invalid
        ],
    )
    def test_empty_counts(self, basis_matrix, trial_counts):
        with pytest.raises(ValueError, match="Empty array provided"):
            convolve.create_convolutional_predictor(basis_matrix, trial_counts, axis=1)

    @pytest.mark.parametrize(
        "time_series, check_func, axis",
        [
            (np.zeros((1, 20)), lambda x: x.ndim == 3, 1),
            (np.zeros((20,)), lambda x: x.ndim == 2, 0),
            ([np.zeros((20,)), np.zeros((10,))], lambda x: x.ndim == 2, 0),
            (np.zeros((20, 1)), lambda x: x.ndim == 3, 0),
            (np.zeros((1, 20, 1)), lambda x: x.ndim == 4, 1),
            ([np.zeros((20, 1)), np.zeros((20, 1))], lambda x: x.ndim == 3, 0),
            ([np.zeros((10, 1)), np.zeros((20, 1))], lambda x: x.ndim == 3, 0),
        ],
    )
    def test_output_ndim(self, time_series, check_func, axis):
        res = convolve.create_convolutional_predictor(
            np.zeros((2, 1)), time_series, axis=axis
        )
        if not utils.pytree_map_and_reduce(check_func, all, res):
            raise ValueError("Output doesn't match expected structure")

    @pytest.mark.parametrize(
        "time_series, axis, output_shape",
        [
            (np.zeros((20, 1)), 0, (20, 1, 1)),
            (np.zeros((1, 20, 1)), 1, (1, 20, 1, 1)),
            ([[np.zeros((1, 1, 20))], np.zeros((1, 1, 20))], 2, (1, 1, 20, 1)),
            ([np.zeros((20, 1))], 0, (20, 1, 1)),
            ([np.zeros((10, 1))], 0, (10, 1, 1)),
            ([[np.zeros((10, 1))]], 0, (10, 1, 1)),
        ],
    )
    def test_output_shape(self, time_series, axis, output_shape):
        def check_func(x):
            return x.shape == output_shape

        res = convolve.create_convolutional_predictor(
            np.zeros((2, 1)), time_series, axis=axis
        )
        if not utils.pytree_map_and_reduce(check_func, all, res):
            raise ValueError("Output  number of neuron doesn't match input.")

    @pytest.mark.parametrize(
        "time_series, axis",
        [
            (np.zeros((20,)), 0),
            (np.zeros((20, 1)), 0),
            (np.zeros((1, 20, 1)), 1),
            ([np.zeros((20, 1)), np.zeros((20, 1))], 0),
            ([np.zeros((10, 1)), np.zeros((20, 1))], 0),
            ([np.zeros((10, 1)), np.zeros((20, 2))], 0),
        ],
    )
    @pytest.mark.parametrize("basis_matrix", [np.zeros((2, 1)), np.zeros((2, 2))])
    def test_output_num_basis(self, time_series, basis_matrix, axis):
        def check_func(conv):
            return basis_matrix.shape[-1] == conv.shape[-1]

        res = convolve.create_convolutional_predictor(
            basis_matrix, time_series, axis=axis
        )
        if not utils.pytree_map_and_reduce(check_func, all, res):
            raise ValueError("Output  number of neuron doesn't match input.")

    @pytest.mark.parametrize(
        "basis_matrix", [np.random.normal(size=(4, 3)) for _ in range(2)]
    )
    @pytest.mark.parametrize(
        "trial_counts", [{key: np.random.normal(size=(10, 3)) for key in range(2)}]
    )
    def test_valid_convolution_output_tree(self, basis_matrix, trial_counts):
        numpy_out = np.zeros(
            (
                len(trial_counts),
                trial_counts[0].shape[0] - basis_matrix.shape[0] + 1,
                trial_counts[0].shape[1],
                basis_matrix.shape[1],
            )
        )
        for tri_i, trial in trial_counts.items():
            for neu_k, vec in enumerate(trial.T):
                for bas_j, basis in enumerate(basis_matrix.T):
                    numpy_out[tri_i, :, neu_k, bas_j] = np.convolve(
                        vec, basis, mode="valid"
                    )
        ws = basis_matrix.shape[0]
        utils_out = convolve.create_convolutional_predictor(
            basis_matrix, trial_counts, axis=0, shift=False
        )
        check = all(
            np.allclose(utils_out[k][ws - 1 :], numpy_out[k], rtol=10**-5, atol=10**-5)
            for k in utils_out
        )
        assert check, (
            "Output of utils.convolve_1d_trials "
            "does not match numpy.convolve in "
            '"valid" mode.'
        )

    @pytest.mark.parametrize(
        "trial_counts, axis",
        [
            (np.zeros((1, 30, 2)), 1),
            ([np.zeros((30, 2))], 0),
            ({"tr1": np.zeros((30, 2)), "tr2": np.zeros((30, 2))}, 0),
            (np.zeros((30, 10)), 0),
            ([np.zeros((30, 10))], 0),
            ({"nested": [{"tr1": np.zeros((30, 2)), "tr2": np.zeros((30, 2))}]}, 0),
        ],
    )
    def test_tree_structure_match(self, trial_counts, axis):
        basis_matrix = np.zeros((4, 3))
        conv = convolve.create_convolutional_predictor(
            basis_matrix, trial_counts, axis=axis
        )
        assert jax.tree_util.tree_structure(
            trial_counts
        ) == jax.tree_util.tree_structure(conv)

    @pytest.mark.parametrize("axis", [0, 1, 2])
    @pytest.mark.parametrize(
        "window_size, shift, predictor_causality, nan_idx",
        [
            (3, True, "causal", [0, 1, 2]),
            (2, True, "causal", [0, 1]),
            (3, False, "causal", [0, 1]),
            (2, False, "causal", [0]),
            (2, None, "causal", [0, 1]),
            (3, True, "anti-causal", [29, 28, 27]),
            (2, True, "anti-causal", [29, 28]),
            (3, False, "anti-causal", [29, 28]),
            (2, False, "anti-causal", [29]),
            (2, None, "anti-causal", [29, 28]),
            (3, False, "acausal", [29, 0]),
        ],
    )
    def test_expected_nan(self, axis, window_size, shift, predictor_causality, nan_idx):
        shape = [1, 1, 1]
        shape[axis] = 30
        feature = np.zeros(shape)
        basis = np.zeros((window_size, 1))
        res = convolve.create_convolutional_predictor(
            basis,
            feature,
            predictor_causality=predictor_causality,
            shift=shift,
            axis=axis,
        )
        # get expected non-nan idxs
        other_idx = list(set(np.arange(res.shape[1])).difference(nan_idx))

        # take values and check
        nans = np.take(res, nan_idx, axis)
        non_nans = np.take(res, other_idx, axis)
        assert np.all(np.isnan(nans))
        assert not np.any(np.isnan(non_nans))

    @pytest.mark.parametrize(
        "tsd",
        [
            nap.Tsd(
                t=np.arange(100),
                d=np.arange(100),
                time_support=nap.IntervalSet(start=[0, 50], end=[20, 75]),
            )
        ],
    )
    @pytest.mark.parametrize(
        "window_size, shift, predictor_causality, nan_index",
        [
            (3, True, "causal", [0, 1, 2, 50, 51, 52]),
            (2, True, "causal", [0, 1, 50, 51]),
            (3, False, "causal", [0, 1, 50, 51]),
            (2, False, "causal", [0, 50]),
            (2, None, "causal", [0, 1, 50, 51]),
            (3, True, "anti-causal", [20, 19, 18, 75, 74, 73]),
            (2, True, "anti-causal", [20, 19, 75, 74]),
            (3, False, "anti-causal", [20, 19, 75, 74]),
            (2, False, "anti-causal", [20, 75]),
            (2, None, "anti-causal", [20, 19, 75, 74]),
            (3, False, "acausal", [0, 20, 50, 75]),
        ],
    )
    def test_multi_epoch_pynapple(
        self, tsd, window_size, shift, predictor_causality, nan_index
    ):
        """Test nan location in multi-epoch pynapple tsd."""
        basis = np.zeros((window_size, 1))
        res = convolve.create_convolutional_predictor(
            basis, tsd, predictor_causality=predictor_causality, shift=shift
        )

        nan_index = np.sort(nan_index)
        times_nan_found = res[np.isnan(res.d[:, 0])].t
        assert len(times_nan_found) == len(nan_index)
        assert all(times_nan_found == np.array(nan_index))



"""Decorator for array casting.

This module provides utilities for seamless type casting between JAX, NumPy, and pynapple time series with data
objects.

It includes functions to check if objects are numpy array-like or pynapple time series data (TSD) objects, verify
consistency in time axes and supports among TSD objects, and perform conditional conversion to JAX arrays.
The key feature for this module is the `cast_jax` decorator, which automatically casts numpy-array like inputs
to JAX arrays and, where applicable, converts outputs back to pynapple TSD objects.
"""

from functools import wraps
from typing import Any, Callable, List, Literal, Optional, Type, Union

import jax
import jax.numpy as jnp
import numpy as np
import pynapple as nap
from numpy.typing import NDArray

from . import tree_utils

_NAP_TIME_PRECISION = 10 ** (-nap.nap_config.time_index_precision)


def is_numpy_array_like(obj) -> bool:
    """
    Check if an object is array-like of at least 1-dimension.

    This function determines if an object has array-like properties but isn't an _BaseTsd.
    An object is considered array-like if it has attributes typically associated with arrays
    (such as `.shape`, `.dtype`, and `.ndim`), supports indexing, and is iterable.

    Parameters
    ----------
    obj : object
        The object to check for array-like properties.

    Returns
    -------
    :
        True if the object is array-like, False otherwise.

    Notes
    -----
    This function uses a combination of checks for attributes (`shape`, `dtype`, `ndim`),
    indexability, and iterability to determine if the given object behaves like an array.
    It is designed to be flexible and work with various types of array-like objects, including
    but not limited to NumPy arrays and JAX arrays. However, it may not be foolproof for all
    possible array-like types or objects that mimic these properties without being suitable for
    numerical operations.

    """
    # if pandas check obj.value
    obj = (
        obj.values
        if all(hasattr(obj, name) for name in ("values", "index"))
        and not is_pynapple_tsd(obj)
        else obj
    )

    # Check for array-like attributes
    has_shape = hasattr(obj, "shape")
    has_dtype = hasattr(obj, "dtype")
    has_ndim = hasattr(obj, "ndim")

    # Check for indexability (try to access the first element)
    try:
        obj[0]
        is_indexable = True
    except (TypeError, IndexError):
        is_indexable = False

    # Check for iterable property
    try:
        iter(obj)
        is_iterable = True
    except TypeError:
        is_iterable = False

    return has_shape and has_dtype and has_ndim and is_indexable and is_iterable


def is_pynapple_tsd(x: Any) -> bool:
    """
    Verify if an object is a pynapple time series with data object.

    Examines the presence of specific attributes that are characteristic of pynapple time series data structures.

    Parameters
    ----------
    x
        Object to evaluate.

    Returns
    -------
    :
        Indicates the result of the evaluation.
    """
    return all(hasattr(x, attr) for attr in ["times", "data", "time_support"])


def _has_same_time_axis(*args, **kwargs) -> bool:
    """
    Check for matching time axes among pynapple objects.

    Evaluates whether the time axes of provided pynapple objects are close enough to be considered identical,
    using a predefined precision level.

    Parameters
    ----------
    args:
        Any positional argument.
    kwargs:
        Any keyword argument.


    Returns
    -------
    :
    Indicates whether all time axes match.
    """
    flat_tree, _ = jax.tree_util.tree_flatten((args, kwargs))

    if len(flat_tree) == 1:
        return True

    # get first pynapple
    is_nap = (is_pynapple_tsd(x) for x in flat_tree)
    time = [x.t for x, bl in zip(flat_tree, is_nap) if bl]

    # check time samples are close (using pynapple precision)
    return _check_all_close(time)


def _has_same_support(*args, **kwargs):
    """
    Verify matching time support intervals among pynapple objects.

    Evaluates whether the time support intervals of provided pynapple objects are close enough to be
    considered identical, using a predefined precision level.

    Parameters
    ----------
    args:
        Any positional argument.
    kwargs:
        Any keyword argument.

    Returns
    -------
    :
        Indicates the result of the verification.
    """
    flat_tree, _ = jax.tree_util.tree_flatten((args, kwargs))

    if len(flat_tree) == 1:
        return True

    # get first pynapple
    is_nap = (is_pynapple_tsd(x) for x in flat_tree)
    time_support = [x.time_support.values for x, bl in zip(flat_tree, is_nap) if bl]

    # check starts and ends are close (using pynapple precision)
    return _check_all_close(time_support)


def _check_all_close(arrays: List[NDArray]) -> bool:
    """
    Check that equality of two arrays up to numerical precision.

    Parameters
    ----------
    arrays:
        List of arrays to compare

    Returns
    -------
    :
        True if the array are equal up to numerical precision, False otherwise.

    """
    if not all(arrays[0].ndim == arr.ndim for arr in arrays[1:]):
        return False

    elif not all(arrays[0].shape == arr.shape for arr in arrays[1:]):
        return False

    return all(
        jnp.allclose(
            arrays[0],
            x,
            rtol=0,
            atol=_NAP_TIME_PRECISION,
        )
        for x in arrays[1:]
    )


def all_same_time_info(*args, **kwargs) -> bool:
    """
    Ensure consistent time axis and support information among pynapple objects.

    Combines checks for matching time axes and support intervals to verify consistency across all provided pynapple
    objects.

    Parameters
    ----------
    args:
        Any positional argument.
    kwargs:
        Any keyword argument.

    Returns
    -------
    :
        Indicates whether time-related information is consistent.
    """
    return _has_same_time_axis(*args, **kwargs) and _has_same_support(*args, **kwargs)


def get_time_info(*args, **kwargs):
    """
    Extract time axis and support information from the first pynapple object.

    Assumes the presence of at least one pynapple object among the inputs and retrieves its time-related information.

    Parameters
    ----------
    args:
        Any positional argument.
    kwargs:
        Any keyword argument.

    Returns
    -------
    :
        Time axis and support information of the first pynapple object detected.
    """
    # list of bool
    flat, _ = jax.tree_util.tree_flatten(
        jax.tree_util.tree_map(is_pynapple_tsd, (args, kwargs))
    )

    idx = flat.index(True)

    # get the corresponding tsd
    tsd, _ = jax.tree_util.tree_flatten((args, kwargs))

    return tsd[idx].t, tsd[idx].time_support


def cast_to_pynapple(
    array: jnp.ndarray, time: NDArray, time_support: nap.IntervalSet
) -> Union[nap.Tsd, nap.TsdFrame, nap.TsdTensor]:
    """
    Convert an array to a pynapple time series object.

    Depending on the dimensionality of the array, creates an appropriate pynapple time series data structure using
    the provided time and support information.

    Parameters
    ----------
    array:
        Data array to convert.
    time:
        Time axis for the pynapple object.
    time_support:
        Time support information for the pynapple object.

    Returns
    -------
    :
    A pynapple time series object based on the input array.
    """
    # keep time on CPU, pynapple numba operations on time are more efficient
    time = np.asarray(time)
    if time.shape[0] != array.shape[0]:
        return array
    elif array.ndim == 1:
        return nap.Tsd(t=time, d=array, time_support=time_support)
    elif array.ndim == 2:
        return nap.TsdFrame(t=time, d=array, time_support=time_support)
    else:
        return nap.TsdTensor(t=time, d=array, time_support=time_support)


def jnp_asarray_if(
    x: Any,
    condition: Callable[[Any], bool] = is_numpy_array_like,
    dtype: Optional[Type] = None,
) -> Any:
    """
    Conditionally convert an object to a JAX array.

    Applies the conversion if the specified condition is met. Allows for flexible handling of inputs that should
    be treated as arrays for numerical computations.

    Parameters
    ----------
    x:
        Object to potentially convert.
    condition:
        A callable that determines whether conversion should occur.
    dtype:
        dtype for the conversion.

    Returns
    -------
    :
        The original object or its conversion to a JAX array, based on the condition.
    """
    if condition(x):
        x = jnp.asarray(x, dtype=dtype)
    return x


def np_asarray_if(
    x: Any, condition: Callable[[Any], bool] = is_numpy_array_like
) -> Any:
    """
    Conditionally convert an object to a numpy array.

    Applies the conversion if the specified condition is met. Allows for flexible handling of inputs that should
    be treated as arrays for numerical computations.

    Parameters
    ----------
    x:
        Object to potentially convert.
    condition:
        A callable that determines whether conversion should occur.

    Returns
    -------
    :
        The original object or its conversion to a numpy array, based on the condition.
    """
    if condition(x):
        x = np.asarray(x)
    return x


def support_pynapple(conv_type: Literal["jax", "numpy"] = "jax") -> Callable:
    """
    Decorate a function to cast inputs between JAX arrays and pynapple objects.

    Automatically converts input arguments to JAX arrays for processing. If the original inputs include pynapple
    objects, attempts to convert the function's output back to pynapple format while ensuring consistency in the
    time axis.

    Parameters
    ----------
    conv_type
        The type of conversion. Either "numpy" or "jax".

    Returns
    -------
    :
        A wrapper function that applies the specified casting behavior.

    Raises
    ------
    NotImplementedError:
        If the conversion type is not implemented.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # check for the presence of any pynapple tsd/tsdFrame/tsdTensor
            any_nap = tree_utils.pytree_map_and_reduce(
                is_pynapple_tsd, any, (args, kwargs)
            )

            # type casting pynapple
            if any_nap:
                # check if the time axis is the same
                if not all_same_time_info(*args, **kwargs):
                    raise ValueError(
                        "Time axis mismatch. pynapple objects have mismatching time axis."
                    )
                time, time_support = get_time_info(*args, **kwargs)

                def cast_out(tree):
                    # cast back to pynapple
                    return jax.tree_util.tree_map(
                        lambda x: cast_to_pynapple(x, time, time_support), tree
                    )

            else:
                # if no pynapple time series is present, apply the function/method
                return func(*args, **kwargs)

            if conv_type == "jax":
                # cast to jax
                args, kwargs = jax.tree_util.tree_map(jnp_asarray_if, (args, kwargs))
            elif conv_type == "numpy":
                # cast to numpy
                args, kwargs = jax.tree_util.tree_map(np_asarray_if, (args, kwargs))
            else:
                raise NotImplementedError(
                    f"Conversion of type '{conv_type}' not implemented!"
                )
            # apply function/method
            res = func(*args, **kwargs)
            # revert casting if pynapple
            return cast_out(res)

        return wrapper

    return decorator
